import numpy as np

#import pyximport; pyximport.install()
from gmtb.util.distance.kendall_tau_dist_cy import tau_cy, tau_matrix_cy, try_bucket_cy

# computes the kendall tau distance between rankings when rankings are encoded as index vectors
# each vector is an array r where r[i] is the bucket of value i.
# Example:
# [1], [2,4], [3,5,6] is encoded as [1, 2, 3, 2, 3, 3]
def kendall_tau_dist(r1, r2, penalty=0.5):
    assert(r1.dtype == np.int)
    assert (r2.dtype == np.int)
    return tau_cy(r1,r2,penalty)



def kendall_tau_wm(r1, r2, alpha, penalty=0.5):
    assert(r1.dtype == np.int)
    assert (r2.dtype == np.int)

    r_med = r1
    done = False

    # matrix where disagreements happen
    mat_r1 = np.zeros((len(r1), len(r1)))
    #mat_r2 = tau_matrix(r_med, r2, penalty)
    mat_r2 = tau_matrix_cy(r_med, r2, penalty)

    # current tau
    tau_r1 = 0
    tau_r2 = np.sum(mat_r2)

    # goal tau
    tau_r1_goal = alpha * tau_r2
    tau_r2_goal = (1 - alpha) * tau_r2

    #iter = 0

    # iterate all possible values and try to find a best possible mean
    while (abs(tau_r1 - tau_r1_goal) > 1e-10 or abs(tau_r2 - tau_r2_goal) > 1e-10) and not done:

        #iter = iter + 1
        #print('Iteration %d, %f %f %f %f' % (iter,tau_r1, tau_r1_goal, tau_r2, tau_r2_goal))

        best_r_med = r_med
        best_tau_r1 = tau_r1
        best_tau_r2 = tau_r2
        best_mat_r1 = mat_r1
        best_mat_r2 = mat_r2
        done = True

        disagreeing_values = np.sum(mat_r1 + mat_r2, axis=1) != 0
        for value in range(len(r_med)):
            # try full bucket
            if disagreeing_values[value]:
                for bucket in range(1,max(r_med)+1):
                    # existing bucket, skip if its already in there
                    if (r_med[value] != bucket):
                        #r_med_new, mat_r1_new, mat_r2_new, tau_r1_new, tau_r2_new = try_bucket(r1 ,r2, r_med, mat_r1, mat_r2, value, bucket, penalty, False)
                        r_med_new, mat_r1_new, mat_r2_new, tau_r1_new, tau_r2_new = try_bucket_cy(r1, r2, r_med, mat_r1,
                                                                                                  mat_r2, value, bucket,
                                                                                                  penalty, False)
                        # check if better
                        if abs(tau_r1_new - tau_r1_goal) + abs(tau_r2_new - tau_r2_goal) < abs(best_tau_r1 - tau_r1_goal) + abs(
                                best_tau_r2 - tau_r2_goal):
                            best_r_med = r_med_new
                            best_mat_r1 = mat_r1_new
                            best_mat_r2 = mat_r2_new
                            best_tau_r1 = tau_r1_new
                            best_tau_r2 = tau_r2_new
                            done = False

                    # new bucket
                    #r_med_new, mat_r1_new, mat_r2_new, tau_r1_new, tau_r2_new = try_bucket(r1 ,r2, r_med, mat_r1, mat_r2, value, bucket, penalty, True)
                    r_med_new, mat_r1_new, mat_r2_new, tau_r1_new, tau_r2_new = try_bucket_cy(r1, r2, r_med, mat_r1,
                                                                                              mat_r2, value, bucket,
                                                                                              penalty, True)
                    if abs(tau_r1_new - tau_r1_goal) + abs(tau_r2_new - tau_r2_goal) < abs(best_tau_r1 - tau_r1_goal) + abs(
                            best_tau_r2 - tau_r2_goal):
                        best_r_med = r_med_new
                        best_mat_r1 = mat_r1_new
                        best_mat_r2 = mat_r2_new
                        best_tau_r1 = tau_r1_new
                        best_tau_r2 = tau_r2_new
                        done = False

                r_med = best_r_med
                tau_r1 = best_tau_r1
                tau_r2 = best_tau_r2
                mat_r1 = best_mat_r1
                mat_r2 = best_mat_r2

    return r_med


def tau_matrix(r1, r2, penalty):
    mat = np.zeros((len(r1), len(r1)))

    for i in range(len(r1)):
        for j in range(len(r1)):
            mat[i,j] = (r1[i] < r1[j] and r2[i] > r2[j] or
                         r1[i] > r1[j] and r2[i] < r2[j]) \
                  + penalty * (r1[i] == r1[j] and r2[i] != r2[j] or
                               r1[i] != r1[j] and r2[i] == r2[j])

    return mat


# computes changes after changing the bucket
def try_bucket(r1, r2, r_med, mat_r1, mat_r2, value, bucket_index, penalty, new_bucket):

    # create copyies so we dont change anything important
    r_med_new = r_med.copy()
    mat_r1 = mat_r1.copy()
    mat_r2 = mat_r2.copy()

    # create new array
    if new_bucket:
        # move all values in bucket new_index
        move_idx = r_med_new >= bucket_index
        r_med_new[move_idx] = r_med_new[move_idx] + 1

    # set new index
    r_med_new[value] = bucket_index

    # update matrices
    for i in range(len(r_med_new)):
        mat_r1[value, i] = (r1[value] < r1[i] and r_med_new[value] > r_med_new[i] or
                            r1[value] > r1[i] and r_med_new[value] < r_med_new[i]) \
                            + penalty * (r1[value] == r1[i] and r_med_new[value] != r_med_new[i] or
                               r1[value] != r1[i] and r_med_new[value] == r_med_new[i])
        mat_r1[i, value] =  mat_r1[value, i]

        mat_r2[value, i] = (r2[value] < r2[i] and r_med_new[value] > r_med_new[i] or
                            r2[value] > r2[i] and r_med_new[value] < r_med_new[i]) \
                            + penalty * (r2[value] == r2[i] and r_med_new[value] != r_med_new[i] or
                               r2[value] != r2[i] and r_med_new[value] == r_med_new[i])
        mat_r2[i, value] = mat_r2[i, value]


    # update tau
    tau_r1 = np.sum(mat_r1)
    tau_r2 = np.sum(mat_r2)

    # return everything
    return r_med_new, mat_r1, mat_r2, tau_r1, tau_r2


# convert list of buckets to index vector
def to_index_vector(r):
    # 1 get max index
    # 2 go through each bucket and save index
    raise NotImplementedError


# convert index vector to list of buckets
def to_list(r):
    r_out = []

    for i in range(1, max(r)):
        r_out.append(r[r == i])



import numpy as np

import scipy.special

# Source:
# Jiao, Yunlong, and Jean-Philippe Vert. "The Kendall and Mallows kernels for permutations."
# IEEE transactions on pattern analysis and machine intelligence 40.7 (2017): 1755-1769.
#
# Based on the mapping:
# phi(r)_ij =  1/sqrt(comb(n,2)) , if r[i] > r[j]
#              0                 , if r[i] == r[j]
#             -1/sqrt(comb(n,2)) , if r[i] < r[j]
# with i < j
def kendall_tau_kernel(r1, r2):
    nc = 0
    nd = 0

    for i in range(len(r1)-1):
        for j in range(i, len(r2)):
            # skip comparing to itself
            if i == j:
                continue

            # sum errors
            if r1[i] < r1[j] and r2[i] < r2[j] or r1[i] > r1[j] and r2[i] > r2[j]:
                nc = nc + 1
            elif r1[i] < r1[j] and r2[i] > r2[j] or r1[i] > r1[j] and r2[i] < r2[j]:
                nd = nd + 1

    return (nc - nd) / scipy.special.comb(len(r1), 2)


import numpy as np
import gmtb.util as util
import gmtb.util.distance as distance
from gmtb.dpe import compute_alpha,best_weighted_mean


# implementation of linear and triangular reconstruction
# returns:
#   rec_obj - the reconstructed object
#   best_value - SOD of the reconstructed object
def linear(object_vector, embedding, object_set, dist_func, weighted_mean_func, n_points):

    # get order of objects by distance
    dist = util.pdist(set1=embedding, set2=[object_vector], func=distance.euclidean_dist)
    ind = np.argsort(dist)

    n = len(object_set)
    iter = min(n, n_points)-1
    best_value = np.full(iter+1, np.inf)
    rec_obj = [None] * (iter+1)
    rec_obj[0] = object_set[ind[0]]
    mean_vec = embedding[ind[0]]

    best_value[0] = np.sum(util.pdist(set1=object_set, set2=[rec_obj[0]], func=dist_func))

    for i in range(iter):
        # get next index

        b = ind[i+1]

        # compute alpha
        alpha = compute_alpha(mean_vec, embedding[b,:], object_vector)

        # compute new weighted mean
        rec_obj[i+1], best_value[i+1] = best_weighted_mean(rec_obj[i], object_set[b], alpha, object_set, dist_func, weighted_mean_func)

    ind = np.argmin(best_value)
    return rec_obj[ind], best_value[ind]


from .partition_dist import partition_dist
from .partition_dist import partition_wm

from .string_edit_dist import string_edit_dist
from .string_edit_dist import string_edit_wm

from .euclidean_dist import euclidean_dist
from .euclidean_dist import euclidean_wm

from .kendall_tau_dist import kendall_tau_dist
from .kendall_tau_dist import kendall_tau_wm



import gmtb.kernel
import numpy as np


def linear(object_set,weights,dist,kernel_func,weighted_mean_func,kernel_matrix,n_points):

    # get order of objects by distance
    ind = np.argsort(dist)

    n = len(object_set)
    iter = min(n,n_points)-1
    best_value_kernel_space = np.full(iter+1, np.inf)
    new_mean = [None] * (iter+1)
    new_mean[0] = object_set[ind[0]]

    k_mean_set = kernel_matrix[:, ind[0]]
    k_mean_mean = kernel_matrix[ind[0], ind[0]]
    best_value_kernel_space[0] = np.sum(np.sqrt(k_mean_mean - 2 * k_mean_set + np.diag(kernel_matrix)))

    for i in range(iter):
        # get next index
        b = ind[i+1]

        # compute alpha
        alpha = gmtb.kernel.compute_alpha(weights, k_mean_set, k_mean_mean, kernel_matrix[:, b], kernel_matrix[b, b], k_mean_set[b])

        # compute new weighted mean
        new_mean[i+1], k_mean_set, k_mean_mean, best_value_kernel_space[i+1] = \
            gmtb.kernel.best_weighted_mean(new_mean[i], object_set[b], alpha, object_set, kernel_matrix, kernel_func, weighted_mean_func)

    ind = np.argmin(best_value_kernel_space)
    return new_mean[ind], best_value_kernel_space[ind]


import numpy as np
from ..compute_alpha import compute_alpha
from ..best_weighted_mean import best_weighted_mean
import gmtb.util
import gmtb.util.distance


# implementation of the linear recursive reconstruction method
# returns:
#   rec_obj - reconstructed object
#   best_value - value of the reconstructed object (at the moment SOD)
def linear_recursive(object_vector, embedding, object_set, dist_func, weighted_mean_func):

    # return if only one object
    if len(object_set) == 1:
        return object_set[0], 0

    # save original set
    orig_set = object_set
    object_set = object_set.copy()

    # first best value: set median
    dist = gmtb.util.pdist(set1=embedding, set2=[object_vector], func=gmtb.util.distance.euclidean_dist)
    ind = np.argmin(dist)
    best_value = np.sum(gmtb.util.pdist(set1=object_set, set2=[object_set[ind]], func=dist_func))
    rec_obj = object_set[ind]

    while len(object_set) > 1:

        dist = gmtb.util.pdist(set1=embedding, set2=[object_vector], func=gmtb.util.distance.euclidean_dist)
        ind = np.argsort(dist)

        new_object_set = []
        new_embedding = []

        for i in range(0, len(object_set)-1,2):

            # compute new weighted mean
            alpha = compute_alpha(embedding[ind[i]], embedding[ind[i+1]], object_vector)

            nm, bv = best_weighted_mean(object_set[ind[i]], object_set[ind[i+1]], alpha, orig_set,
                                        dist_func, weighted_mean_func)

            new_object_set.append(nm)
            new_embedding.append(alpha*embedding[ind[i]] + (1-alpha)*embedding[ind[i+1]])

            # save best
            if bv < best_value:
                best_value = bv
                rec_obj = nm

        # if leftover objects: just copy
        if np.remainder(len(object_set), 2) == 1:
            new_object_set.append(object_set[ind[-1]])
            new_embedding.append(embedding[ind[-1]])

        # prepare next iteration
        object_set = new_object_set
        embedding = new_embedding

    # end: return best result
    return rec_obj, best_value




import numpy as np
from scipy.optimize import linprog
from scipy.special import comb
import scipy.sparse
from .k_medians import k_medians

import networkx as nx

from gmtb.kernel.compute_weights import compute_weights
from gmtb.kernel.kernel_functions import dist_sub_batch


# linear programming based lower bound
def lp_lower_bound(dist_mat):
    n = np.size(dist_mat,1)

    index = 0
    c = np.ones(n)
    #a = np.zeros((3 * comb(n, 2, exact=True) + n, n))
    b = np.zeros((3 * comb(n, 2, exact=True) + n, 1))

    a = scipy.sparse.lil_matrix((3 * comb(n, 2, exact=True) + n, n))
    #b = scipy.sparse.lil_matrix((3 * comb(n, 2, exact=True) + n, 1))

    # for i < j
    # -x_i + -x_j <= -dist_mat_ij
    # -x_i +  x_j <=  dist_mat_ij
    #  x_i + -x_j <=  dist_mat_ij
    for j in range(1,n):
        for i in range(j):
            a[index:index+3, [i, j]] = [[-1, -1], [-1, 1], [1, -1]]
            b[index:index+3] = [[-dist_mat[i, j]], [dist_mat[i, j]], [dist_mat[i, j]]]
            index = index + 3

    # for i:
    # -x_j <= 0
    a[-n:, :] = -np.eye(n)
    for j in range(n):
        a[-n+j, j] = -1
    #b[-n:, :] = np.zeros((n ,1))

    # solve linprog. interior-point seems better in first tests
    result = linprog(c, a, b, options={'sparse': True})

    return np.sum(result.x)

def graph_lower_bound(dist_mat):
    # cost matrix: negative distance and Diagonal is infinity
    #cost_matrix = -dist_mat
    #np.fill_diagonal(cost_matrix, np.inf)

    # get pairs by graph matching
    g = nx.from_numpy_matrix(dist_mat)
    pairs = nx.max_weight_matching(g)

    # sum pair values
    sum = 0
    for (i,j) in pairs:
        sum = sum + dist_mat[i,j]

    return sum


# kernel based lower bound
def kernel_lower_bound(dist_mat,kernel_mat):

    # based on k_lin
    if kernel_mat is None:
        origin_idx = k_medians(dist_mat,1)
        kernel_mat = dist_sub_batch(dist_mat,"lin",origin_idx)
        # based on k_nd
        # kernel_mat = dist_sub_batch(dist_mat, "nd", param=2)


    # compute weights
    weights, k_median_set, k_median_median = compute_weights(kernel_mat, True)

    tmp_bound = k_median_median - 2 * k_median_set + np.diag(kernel_mat)

    # remove too small values
    tmp_bound[tmp_bound < 1e-10] = 1e-10

    return np.sum(np.sqrt(tmp_bound))


def lower_bound(dist_mat, method="lp", kernel_mat=None):
    if method == "lp":
        return lp_lower_bound(dist_mat)
    elif method == "graph":
        return graph_lower_bound(dist_mat)
    elif method == "kernel":
        return kernel_lower_bound(dist_mat,kernel_mat)
    else:
        raise ValueError("lower_bound: Wrong method")


from pyclustering.cluster.kmedoids import kmedoids
import numpy as np


def k_medians(dist_mat, subset_size):
    # create K-Medoids algorithm for processing distance matrix instead of points
    initial_medoids = range(subset_size)
    kmedoids_instance = kmedoids(dist_mat, initial_medoids, data_type='distance_matrix')

    # run cluster analysis and obtain results
    kmedoids_instance.process()

    #clusters = kmedoids_instance.get_clusters()
    medoids = np.asarray(kmedoids_instance.get_medoids(),dtype=np.int64)

    return medoids



import numpy as np


# compute_alpha computes the alpha of the projection in kernel space
# weights = weights of the objects
# k_a_set = array of kernel values between object a and the set
# k_a_a = kernel between object a and itself
# k_b_set = array of kernel values between object b and the set
# k_b_b = kernel between object b and itself
# k_a_b = kernel between object a and b
def compute_alpha(weights, k_a_set, k_a_a, k_b_set, k_b_b, k_a_b):

    div = k_b_b - 2 * k_a_b + k_a_a

    # in case of negative values in sqrt, return 0.5 instead
    if div > 0:
        alpha = (np.sum(weights * (k_b_set - k_a_set)) / np.sum(weights) - k_a_b + k_a_a) \
                / np.power(np.sqrt(div), 2)
    else:
        alpha = 0.5

    alpha = max(alpha, 0)
    alpha = min(alpha, 1)
    return alpha



import gmtb.dpe.embedding
from gmtb.util import pdist
import numpy as np
import sklearn.manifold


def embed(object_set, embedding_type, num_dim, dist_func, dist_mat = None):

    # check input
    num_dim = round(num_dim)

    # do embedding
    if embedding_type == "fastmap":
        #TODO: testing
        if dist_mat is None:
            dist_mat = pdist(object_set, dist_func)
        emb = gmtb.dpe.embedding.fastmap(dist_mat.copy(), num_dim)

    elif embedding_type == "metricmap":
        emb = gmtb.dpe.embedding.metricmap(object_set,dist_func,num_dim,dist_mat)

    elif embedding_type == "sparsemap":
        emb = gmtb.dpe.embedding.sparsemap(object_set,dist_func,num_dim,dist_mat)

    elif embedding_type == "mds":
        if dist_mat is None:
            dist_mat = pdist(object_set, dist_func)
        mds = sklearn.manifold.MDS(n_components=num_dim,metric=False,dissimilarity='precomputed')
        emb = mds.fit_transform(dist_mat)

    elif embedding_type == "sammon":
        # TODO: remove duplicates
        if dist_mat is None:
            dist_mat = pdist(object_set, dist_func)
        emb, _ = gmtb.dpe.embedding.sammon(dist_mat, num_dim, inputdist='distance', init='random', display=0)

    elif embedding_type == "cca":
        if dist_mat is None:
            dist_mat = pdist(object_set, dist_func)
        emb = gmtb.dpe.embedding.cca(dist_mat, num_dim, dist_mat=dist_mat)

    elif embedding_type == "t-sne":
        if dist_mat is None:
            dist_mat = pdist(object_set, dist_func)
        tsne = sklearn.manifold.TSNE(n_components=num_dim,metric='precomputed',method='exact')
        emb = tsne.fit_transform(dist_mat)

    #TODO: Isomap,LLE from sklearn.manifold?

    elif embedding_type == "prototype":
        if dist_mat is None:
            dist_mat = pdist(object_set, dist_func)
        emb = gmtb.dpe.embedding.prototype(dist_mat, num_dim)

    else:
        raise NotImplementedError("Embedding Method not implemented")

    return emb




import numpy as np
from gmtb.util import pdist
import networkx as nx


def ewm_get_pairs(dist_mat):
    # cost matrix: negative distance and Diagonal is infinity
    #cost_matrix = -dist_mat
    #np.fill_diagonal(cost_matrix,np.inf)
    #assignment = munkres(cost_matrix).nonzero()

    g = nx.from_numpy_matrix(dist_mat)
    pairs = nx.max_weight_matching(g)

    return pairs


def ewm(dataset, dist_func, weighted_mean_func, verbose=False):
    # EWM calculates the generalized median of objects using an evolutionary weighted mean approach
    # Published in "Evolutionary Weighted Mean Based Framework for Generalized Median Computation with
    # Application to Strings"
    # by Lucas Franek and Xiaoyi Jiang
    #
    # Parameters:
    #   dataset             - Cell array of objects.
    #   dist_func           - Function handle @(o1,o2). Distance function between objects o1 and o2. Objects are
    #                         single cells from dataset
    #   weighted_mean_func  - Function handle @(o1,o2,alpha). Weighted mean function between objects. o1, o2 are
    #                         single cells, alpha is a float
    #                         between 0 and 1.
    #
    # Returns:
    #   best_object         - Resulting best object according to best_crit
    #   best_crit_value     - Sum of Distance of the best object

    # check if there is only one object
    if len(dataset) == 1:
        return dataset[0], 0

    # initialize sets. Copy to not change anything
    original_set = dataset.copy()
    dataset = dataset.copy()
    crit_values = np.sum(pdist(dataset,dist_func),axis=1).tolist()
    last_best_crit = np.inf  # max because min will be found later

    # ********************************* Parameters ********************************* #
    w = 3
    max_iter = 20
    n_max = min(20,2*len(dataset))

    # ********************************* BEGIN ITERATION ********************************* #
    iteration = 0
    stop = False
    stop_again = False

    while iteration  < max_iter and ~stop:

        iteration = iteration+1
        resSOD = []
        n = len(dataset)

        # compute Matrix of distances / similarities
        dist_mat = pdist(dataset,dist_func)

        # get pairs of objects
        pairs = ewm_get_pairs(dist_mat)

        # ********* iterate over pairs of objects *********
        for (x,y) in pairs:

            ## find best alpha for the weighted mean using linear search
            #search_crit = @(alpha) best_crit(original_set, weighted_mean_func(dataset{pairs(i,1)}, dataset{pairs(i,2)}, alpha));
            #options.TolX = last_best_crit * 0.001;
            #[alpha, ~] = fminbnd(search_crit,0,1,options);

            # use this alpha to calculate a result
            #dataset{n+i} = weighted_mean_func(dataset{pairs(i,1)}, dataset{pairs(i,2)}, alpha);
            #crit_values(n+i) = best_crit(original_set,dataset{n+i});

            ## use equidistant alphas
            for step in range(1,w+1):
                alpha = step/(1+w)
                new_obj = weighted_mean_func(dataset[x], dataset[y], alpha)
                dataset.append(new_obj)
                crit_values.append(np.sum(pdist(set1=original_set,set2=[new_obj],func=dist_func)))

        ## ********* delete from result set *********

        idx = np.argsort(crit_values)

        max_length = min(len(dataset),n_max)

        dataset = [dataset[i] for i in idx[:max_length]]
        crit_values = [crit_values[i] for i in idx[:max_length]]


        ## stopping criterium: no change for 2 iterations
        if abs(last_best_crit - crit_values[0]) <= 1e-10:
            if stop_again:
                stop = True
            else:
                stop_again = True

        else:
            last_best_crit = crit_values[0]
            stop_again = False



    # ********************************* END ITERATION ********************************* #
    if verbose:
        print('EWM terminated after %d / %d iterations' % (iteration, max_iter))

    return dataset[0], crit_values[0]


cimport numpy as np
import numpy as np

# this file uses cython to speedup certain calculations for the kendall tau distance and weighted mean


# computes tau itself
def tau_cy(np.ndarray[np.int_t, ndim=1] r1, np.ndarray[np.int_t, ndim=1] r2, double penalty):

    cdef double tau = 0

    # create view
    cdef int r1_len = r1.size



    for i in range(r1_len):
        for j in range(r1_len):
            tau = tau +  (r1[i] < r1[j] and r2[i] > r2[j] or
                         r1[i] > r1[j] and r2[i] < r2[j]) \
                  + penalty * (r1[i] == r1[j] and r2[i] != r2[j] or
                               r1[i] != r1[j] and r2[i] == r2[j])

    return tau

# computes the tau-matrix
def tau_matrix_cy(np.ndarray[np.int_t, ndim=1] r1, np.ndarray[np.int_t, ndim=1] r2, double penalty):
    mat = np.zeros((len(r1), len(r1)))

    # create view
    cdef np.ndarray[np.double_t, ndim=2] mat_view = mat
    cdef int r1_len = r1.size



    for i in range(r1_len):
        for j in range(r1_len):
            mat_view[i,j] = (r1[i] < r1[j] and r2[i] > r2[j] or
                         r1[i] > r1[j] and r2[i] < r2[j]) \
                  + penalty * (r1[i] == r1[j] and r2[i] != r2[j] or
                               r1[i] != r1[j] and r2[i] == r2[j])

    return mat


# computes changes after changing the bucket
def try_bucket_cy(np.ndarray[np.int_t, ndim=1] r1, np.ndarray[np.int_t, ndim=1] r2, np.ndarray[np.int_t, ndim=1] r_med,
                  np.ndarray[np.double_t, ndim=2] mat_r1_py, np.ndarray[np.double_t, ndim=2] mat_r2_py, int value, int bucket_index, double penalty, new_bucket):
#     # create copies so we dont change anything important
    r_med_new_py = r_med.copy()
    mat_r1_py = mat_r1_py.copy()
    mat_r2_py = mat_r2_py.copy()

    # setup variables and views
    cdef int r_med_new_len = len(r_med_new_py)
    cdef np.ndarray[np.int_t, ndim=1] r_med_new = r_med_new_py
    cdef np.ndarray[np.double_t, ndim=2] mat_r1 = mat_r1_py
    cdef np.ndarray[np.double_t, ndim=2] mat_r2 = mat_r2_py

    # create new array
    if new_bucket:
        # move all values in bucket new_index
        move_idx = r_med_new >= bucket_index
        r_med_new[move_idx] = r_med_new[move_idx] + 1

    # set new index
    r_med_new[value] = bucket_index

    # update matrices
    for i in range(r_med_new_len):
        mat_r1[value, i] = (r1[value] < r1[i] and r_med_new[value] > r_med_new[i] or
                            r1[value] > r1[i] and r_med_new[value] < r_med_new[i]) \
                            + penalty * (r1[value] == r1[i] and r_med_new[value] != r_med_new[i] or
                               r1[value] != r1[i] and r_med_new[value] == r_med_new[i])
        mat_r1[i, value] =  mat_r1[value, i]
        mat_r2[value, i] = (r2[value] < r2[i] and r_med_new[value] > r_med_new[i] or
                            r2[value] > r2[i] and r_med_new[value] < r_med_new[i]) \
                            + penalty * (r2[value] == r2[i] and r_med_new[value] != r_med_new[i] or
                               r2[value] != r2[i] and r_med_new[value] == r_med_new[i])
        mat_r2[i, value] = mat_r2[i, value]

    # update tau
    tau_r1 = np.sum(mat_r1)
    tau_r2 = np.sum(mat_r2)

    # return everything
    return r_med_new_py, mat_r1_py, mat_r2_py, tau_r1, tau_r2





import gmtb.util as util


# implementation of prototype embedding with the condition that dist_mat is already computed
def prototype(dist_mat, n_dim):

    prototype_index = util.k_medians(dist_mat, n_dim)

    return dist_mat[:, prototype_index]



import gmtb.dpe


def compute_median(object_set, dist_func, weighted_mean_func, emb_type, num_dim, rec_type):

    emb = gmtb.dpe.embed(object_set, emb_type, num_dim, dist_func)
    median_vector = gmtb.dpe.compute_median_vector(emb)
    median, sod = gmtb.dpe.reconstruct(median_vector, emb, object_set, dist_func, weighted_mean_func, rec_type)

    return median, sod



from .dist_sub import dist_sub
from .dist_sub import dist_sub_batch

from .edit_distance_kernel import edit_kernel
from .edit_distance_kernel import edit_kernel_batch

from .partition_kernel import partition_kernel

from .kendall_kernel import kendall_tau_kernel


import numpy as np
import scipy.optimize
import gmtb.util


# implementation of the linear search method for improving of reconstructions.
# returns
#   rec_obj - the improved reconstructed object
#   best_value - value of the reconstructed object (ath the moment SOD)
def linear_search(median_obj,object_set,dist_func,weighted_mean_func,num_iterations=5):

    best_value = np.sum(gmtb.util.pdist(set1=object_set,set2=[median_obj],func=dist_func))
    rec_obj = median_obj
    best_obj = median_obj

    # for each iteration
    for iter in range(num_iterations):

        last_best_value = best_value

        # try each object
        for obj in range(len(object_set)):

            def search_crit(alpha):
                wm = weighted_mean_func(rec_obj, object_set[obj], alpha)
                return np.sum(gmtb.util.pdist(set1=object_set, set2=[wm], func=dist_func))

            alpha, new_best_value, ierr, numfunc = scipy.optimize.fminbound(search_crit, 0, 1, maxfun=15, full_output=True, disp=0)

            if new_best_value < best_value:
                best_value = new_best_value
                best_obj = weighted_mean_func(rec_obj, object_set[obj], alpha)

        # set new median
        rec_obj = best_obj

        if last_best_value == best_value:
            return rec_obj, best_value

    return rec_obj, best_value


import numpy as np


# projects x onto the line between a and b, returns the alpha between them
def compute_alpha(a, b, x):
    v = x-a
    u = b-a

    alpha = 0.5

    # check for division by 0
    div = np.dot(u, u)
    if div != 0:
        alpha = np.dot(u, v) / np.dot(u, u)
        alpha = max(alpha, 0)
        alpha = min(alpha, 1)

    return alpha



import numpy as np
import gmtb.util as util
from gmtb.util.distance.euclidean_dist import euclidean_dist

# code base from SOM Toolbox 2.1


def potency_curve(v0, vn, l):
    return v0 * np.power(vn/v0, np.arange(0, l, 1) / l)


def cca(data, num_dim, projection=None, epochs=1000, dist_mat=None, alpha0=0.5, lambda0=None):

    # CCA Projects data vectors using Curvilinear Component Analysis.
    # P = cca(D, projection, epochs, [Dist], [alpha0], [lambda0])
    #
    # P = cca(D, 2, 10); % projects the given data to a plane
    # P = cca(D, pcaproj(D, 2), 5); % same, but with PCA initialization
    # P = cca(D, 2, 10, Dist); % same, but the given distance matrix is used
    #
    # Input and output arguments([] 's are optional):
    # D          (matrix) the data matrix, size dlen x dim
    #            (struct) data or map struct
    # projection (scalar) output dimension
    #            (matrix) size dlen x odim, the initial projection
    # epochs     (scalar) training length
    # [Dist]     (matrix) pairwise distance matrix, size dlen x dlen.
    #                     If the distances in the input space should
    #                     be calculated otherwise than as euclidian
    #                     distances, the distance from each vector
    #                     to each other vector can be given here,
    #                     size dlen x dlen.For example PDIST
    #                     function can be used to calculate the
    #                     distances: Dist = squareform(pdist(D, 'mahal'));
    # [alpha0]   (scalar) initial step size, 0.5 by default
    # [lambda0]  (scalar) initial radius of influence, 3 * max(std(D)) by default
    #
    # Output:
    # projection (matrix) size dlen x odim, the projections
    #
    # Unknown values(NaN's) in the data: projections of vectors with
    # unknown components tend to drift towards the center of the
    # projection distribution.Projections of totally unknown vectors are
    # set to unknown(NaN).
    #
    # See also SAMMON, PCAPROJ.

    # Reference: Demartines, P., Herault, J., "Curvilinear Component
    #      Analysis: a Self - Organizing Neural Network for Nonlinear
    #      Mapping of Data Sets", IEEE Transactions on Neural Networks,
    #      vol 8, no 1, 1997, pp.148-154.

    # Contributed to SOM Toolbox 2.0, February 2nd, 2000 by Juha Vesanto
    # Copyright (c) by Juha Vesanto
    # http://www.cis.hut.fi/projects/somtoolbox/

    #juuso 171297 040100


    # input data
    noc = np.shape(data)[0]

    me = np.mean(data, axis=1)
    st = np.std(data, axis=1)

    #initial projection
    if projection is None:
        projection = (2 * np.random.rand(noc, num_dim) - 1) * st[0:num_dim] + me[0: num_dim]

    else:
        # replace unknown projections with known values
        inds = np.where(np.isnan(projection))
        projection[inds] = np.random.rand(np.shape(inds))

    # training length
    train_len = epochs * noc

    # random sample order
    # rand('state', sum(100 * clock)); % % TODO: FIX
    sample_inds = np.ceil(noc * np.random.rand(train_len)).astype(np.int64)-1 # -1 because index is from -0 to noc-1

    # mutual distances
    if dist_mat is None or np.alltrue(np.isnan(dist_mat)):
        # fprintf(2, 'computing mutual distances\r');
        #TODO compute euclidean distances
        raise NotImplementedError("Only with Distance matrix so far")
    else:
        # if there is a distance matrix
        if np.shape(dist_mat)[0] != noc:
            raise ValueError('Mutual distance matrix size and data set size do not match')


    # alpha and lambda
    alpha = potency_curve(alpha0, alpha0 / 100, train_len)

    if lambda0 is None:
        lambda0 = np.max(st) * 3

    lambdas = potency_curve(lambda0, 0.01, train_len)


    # ------------------- Action ------------------------

    k = 0 #fprintf(2, 'iterating: %d / %d epochs\r', k, epochs);

    for i in range(train_len):
        ind = sample_inds[i]  # sample index
        dx = dist_mat[:, ind]  # mutual distances in input space
        known = np.nonzero(np.logical_not(np.isnan(dx)))  # known distances

        if np.size(known) > 0:
            # sample vector's projection
            y = projection[ind, :]

            # distances in output space
            Dy = np.squeeze(projection[known, :] - y)
            dy = np.sqrt(np.sum(np.power(Dy,2),axis=1))

            # relative effect
            dy[dy == 0] = 1  # to get rid of div - by - zero's
            fy = np.exp(-dy / lambdas[i]) * (dx[known] / dy - 1)

            # Note that the function F here is e ^ (-dy /lambda))
            # instead of the bubble function 1(lambda -dy) used in the
            # paper.

            # Note that here a simplification has been made: the derivatives of the
            # F function have been ignored in calculating the gradient of error
            # function w.r.t.to changes in dy.

            # update
            projection[known, :] = np.squeeze(projection[known, :]) + alpha[i] * fy[:, np.newaxis] * Dy


        # track
        if np.remainder(i, noc) == 0:
            k = k + 1  # fprintf(2, 'iterating: %d / %d epochs\r', k, epochs);


    # -----------  Clear up -------------

    # calculate error
    # c_error = cca_error(P, Mdist, lambda (train_len));
    # fprintf(2, '%d iterations, error %f          \n', epochs, c_error);

    # set projections of totally unknown vectors as unknown
    unknown = np.sum(np.isnan(data), axis=1) == num_dim
    projection[unknown, :] = np.NaN

    return projection



import numpy as np
import editdistance


# compute table and pointer matrix
def string_edit_table(s1, s2):
    m = len(s1) + 1
    n = len(s2) + 1

    # init
    tbl = np.zeros((m, n))
    for i in range(m): tbl[i, 0] = i
    for j in range(n): tbl[0, j] = j

    # fill table
    for i in range(1, m):
        for j in range(1, n):
            cost = 0 if s1[i - 1] == s2[j - 1] else 1
            tbl[i, j] = min(tbl[i    , j - 1] + 1,
                            tbl[i - 1, j    ] + 1,
                            tbl[i - 1, j - 1] + cost)

    return tbl


# computes the partition distance between two clusterings
def string_edit_dist(s1, s2):
    return editdistance.eval(s1,s2)
    #return string_edit_table(s1, s2)[len(s1), len(s2)]


# computes the weighted mean between two strings
def string_edit_wm(s1, s2, alpha):
    tbl = string_edit_table(s1, s2)

    # go backwards for pointers
    step_list = []

    i = len(s1)
    j = len(s2)

    distance = tbl[i, j]

    while i > 0 and j > 0:
        if tbl[i, j] == tbl[i, j - 1] + 1:  # ins
            step_list.append("ins")
            j = j - 1
        elif tbl[i, j] == tbl[i - 1, j] + 1:  # del
            step_list.append("del")
            i = i - 1
        else:   # sub
            step_list.append("sub")
            j = j - 1
            i = i - 1

    while i > 0:
        step_list.append("del")
        i = i - 1

    while j > 0:
        step_list.append("ins")
        j = j - 1

    # go through list for actions
    num_steps = round(alpha*distance)
    wm_string = ""
    s = 0
    i = 0
    j = 0
    while s < num_steps:
        step = step_list.pop()

        if step == "ins":        # ins
            wm_string = wm_string + s2[j]
            j = j + 1
            s = s + 1
        elif step == "del":      #
            i = i + 1
            s = s + 1
        else:                    # sub
            wm_string = wm_string + s2[j]
            if s1[i] != s2[j]:
                s = s + 1
            i = i + 1
            j = j + 1

    if i < len(s1):
        wm_string = wm_string + s1[i:len(s1)]

    return wm_string



# test for sting wm
if __name__ == '__main__':
    print(string_edit_wm("BBAABB", "AA", 1))


import gmtb.util as util
import numpy as np


def best_weighted_mean(oa, ob, alpha, object_set, dist_func, weighted_mean_func):
    mean1 = weighted_mean_func(oa, ob, alpha)
    mean2 = weighted_mean_func(ob, oa, 1-alpha)

    best_value1 = np.sum(util.pdist(set1=object_set, set2=[mean1], func=dist_func))
    best_value2 = np.sum(util.pdist(set1=object_set, set2=[mean2], func=dist_func))

    if best_value1 < best_value2:
        return mean1, best_value1
    else:
        return mean2, best_value2



import numpy as np
from ..compute_alpha import compute_alpha
from ..best_weighted_mean import best_weighted_mean


# implementation of the linear recursive reconstruction method
# returns:
#   rec_obj - reconstructed object
#   best_value - value of the reconstructed object (at the moment SOD)
def linear_recursive(object_set, weights, dist, kernel_func, weighted_mean_func, kernel_matrix):

    # return if only one object
    if len(object_set) == 1:
        return object_set[0], 0

    # save original set
    orig_set = object_set.copy()
    object_set = object_set.copy()

    # arrays for intermediate kernel values
    k_mean_set = kernel_matrix
    k_mean_mean = np.diag(kernel_matrix)

    # first best value: set median
    ind = np.argsort(dist)
    best_value = np.sum(np.sqrt(k_mean_mean[ind[0]] - 2 * k_mean_set[ind[0], :] + np.diag(kernel_matrix)))
    rec_obj = object_set[ind[0]]

    # precompute k_median
    k_median_vec = np.sum(np.outer(weights,weights) * kernel_matrix) / np.sum(weights)**2

    while len(object_set) > 1:

        ind = np.argsort(dist)

        new_object_set = []
        new_k_mean_set = []
        new_k_mean_mean = []
        new_dist = []

        for i in range(0,len(object_set)-1,2):

            # compute new weighted mean
            alpha = compute_alpha(weights, k_mean_set[ind[i]], k_mean_mean[ind[i]], k_mean_set[ind[i+1]], k_mean_mean[ind[i+1]],
                                  kernel_func(object_set[ind[i]],object_set[ind[i+1]]))

            nm, kms, kmm, bv = \
                best_weighted_mean(object_set[ind[i]], object_set[ind[i+1]], alpha, orig_set, kernel_matrix, kernel_func,
                                   weighted_mean_func)

            new_object_set.append(nm)
            new_k_mean_set.append(kms)
            new_k_mean_mean.append(kmm)
            tmp_dist = k_median_vec\
                       - 2 * np.sum(weights * kms) / np.sum(weights)\
                       + kmm

            if tmp_dist < 0:
                tmp_dist = (dist[ind[i]] + dist[ind[i+1]])/2
            else:
                tmp_dist = np.sqrt(tmp_dist)
            new_dist.append(tmp_dist)

            # save best
            if bv < best_value:
                best_value = bv
                rec_obj = nm

        # if leftover objects: just copy
        if np.remainder(len(object_set), 2) == 1:
            new_object_set.append(object_set[-1])
            new_k_mean_mean.append(k_mean_mean[-1])
            new_k_mean_set.append(k_mean_set[-1])
            new_dist.append(dist[-1])

        # prepare next iteration
        object_set = new_object_set
        k_mean_set = new_k_mean_set
        k_mean_mean = new_k_mean_mean
        dist = new_dist

    # end: return best result
    return rec_obj, best_value




import numpy as np


def median_weights_iter(n, kernel_matrix, start, normalize):
    # return values
    weights_median_set = np.full(n, np.inf)
    weights_median_set_new = start
    k_median_set = None
    k_median_set_new = np.zeros(n)
    k_median_median = None
    k_median_median_new = 0

    # stop conditions
    negative_stop = False
    threshold = 1e-15
    max_iter = 1000
    iter = 0

    while np.sum(np.power(weights_median_set - weights_median_set_new, 2)) > threshold and iter < max_iter:
        # if and ~isreal ...

        weights_median_set = weights_median_set_new
        k_median_median = k_median_median_new
        k_median_set = k_median_set_new

        # kernel between median and median
        k_median_median_new = np.sum(kernel_matrix * np.outer(weights_median_set, weights_median_set)) / \
                              np.power(np.sum(weights_median_set), 2)

        # kernel between objects and median
        for i in range(n):
            k_median_set_new[i] = np.sum(kernel_matrix[:, i] * weights_median_set) / np.sum(weights_median_set)

        # compute weights
        # prevent negative values
        tmp_weights = np.diag(kernel_matrix) - 2 * k_median_set_new + k_median_median_new
        tmp_weights[tmp_weights < 1e-10] = 1e-10

        weights_median_set_new = 1 / np.sqrt(tmp_weights)
        # weights_median_set_new[weights_median_set_new < 1e-5] = 1e-5
        # weights_median_set_new = np.sqrt(weights_median_set_new)

        if normalize:
            weights_median_set_new = weights_median_set_new / np.sum(weights_median_set_new)


        iter = iter + 1

    if np.any(np.isnan(weights_median_set_new)):
        negative_stop = True

    return weights_median_set_new, k_median_set_new, k_median_median_new, negative_stop


def compute_weights(kernel_matrix, normalize=True):
    n = np.shape(kernel_matrix)[0]

    start = np.sum(kernel_matrix) / np.power(n, 2) - (2 / n * np.sum(kernel_matrix, 1)) + np.diag(kernel_matrix)
    start[start < 1e-5] = 1e-5
    start = np.sqrt(start)
    start = start.astype(np.float128)

    max_iter = 5
    iter = 0
    weights_median_set = None
    k_median_set = None
    k_median_median = None
    negative_stop = True

    # as long as iteration not stopped and there was no stop because of negative values
    while iter < max_iter and negative_stop:
        weights_median_set, k_median_set, k_median_median, negative_stop = \
            median_weights_iter(n, kernel_matrix, start, normalize)

        start = np.random.rand(n) + 1
        start = start.astype(np.float128)
        iter = iter + 1

    return weights_median_set, k_median_set, k_median_median



import numpy as np


def best_weighted_mean(oa, ob, alpha, object_set, kernel_matrix, kernel_func, weighted_mean_func):

    n = len(object_set)

    # compute weighted median
    new_mean1 = weighted_mean_func(oa, ob, alpha)
    new_mean2 = weighted_mean_func(ob, oa, 1-alpha)

    # compute sets
    k_mean_set1 = np.zeros(n)
    k_mean_set2 = np.zeros(n)

    for i in range(n):
        k_mean_set1[i] = kernel_func(object_set[i], new_mean1)
        k_mean_set2[i] = kernel_func(object_set[i], new_mean2)

    k_mean_mean1 = kernel_func(new_mean1, new_mean1)
    k_mean_mean2 = kernel_func(new_mean2, new_mean2)

    # compute best value, ignore negative values
    best_value1 = k_mean_mean1 - 2 * k_mean_set1 + np.diag(kernel_matrix)
    best_value2 = k_mean_mean2 - 2 * k_mean_set2 + np.diag(kernel_matrix)
    best_value1[best_value1 < 0] = 0
    best_value2[best_value2 < 0] = 0
    best_value1 = np.sum(np.sqrt(best_value1))
    best_value2 = np.sum(np.sqrt(best_value2))

    # return better result
    if best_value1 < best_value2:
        return new_mean1, k_mean_set1, k_mean_mean1, best_value1
    else:
        return new_mean2, k_mean_set2, k_mean_mean2, best_value2



import numpy as np
from scipy.spatial.distance import cdist


def cmdscale(D):
    """
    Classical multidimensional scaling (MDS)

    Parameters
    ----------
    D : (n, n) array
        Symmetric distance matrix.

    Returns
    -------
    Y : (n, p) array
        Configuration matrix. Each column represents a dimension. Only the
        p dimensions corresponding to positive eigenvalues of B are returned.
        Note that each dimension is only determined up to an overall sign,
        corresponding to a reflection.

    e : (n,) array
        Eigenvalues of B.

    """
    # Number of points
    n = len(D)

    # Centering matrix
    H = np.eye(n) - np.ones((n, n)) / n

    # YY^T
    B = -H.dot(D ** 2).dot(H) / 2

    # Diagonalize
    evals, evecs = np.linalg.eigh(B)

    # Sort by eigenvalue in descending order
    idx = np.argsort(evals)[::-1]
    evals = evals[idx]
    evecs = evecs[:, idx]

    # Compute the coordinates using positive-eigenvalued components only
    w, = np.where(evals > 0)
    L = np.diag(np.sqrt(evals[w]))
    V = evecs[:, w]
    Y = V.dot(L)


    return Y, evals[evals > 0]


def sammon(x, n, display=2, inputdist='raw', maxhalves=20, maxiter=500, tolfun=1e-9, init='default'):
    """Perform Sammon mapping on dataset x
    y = sammon(x) applies the Sammon nonlinear mapping procedure on
    multivariate data x, where each row represents a pattern and each column
    represents a feature.  On completion, y contains the corresponding
    co-ordinates of each point on the map.  By default, a two-dimensional
    map is created.  Note if x contains any duplicated rows, SAMMON will
    fail (ungracefully). 
    [y,E] = sammon(x) also returns the value of the cost function in E (i.e.
    the stress of the mapping).
    An N-dimensional output map is generated by y = sammon(x,n) .
    A set of optimisation options can be specified using optional
    arguments, y = sammon(x,n,[OPTS]):
       maxiter        - maximum number of iterations
       tolfun         - relative tolerance on objective function
       maxhalves      - maximum number of step halvings
       input          - {'raw','distance'} if set to 'distance', X is 
                        interpreted as a matrix of pairwise distances.
       display        - 0 to 2. 0 least verbose, 2 max verbose.
       init           - {'pca', 'cmdscale', random', 'default'}
                        default is 'pca' if input is 'raw', 
                        'msdcale' if input is 'distance'
    The default options are retrieved by calling sammon(x) with no
    parameters.
    File        : sammon.py
    Date        : 18 April 2014
    Authors     : Tom J. Pollard (tom.pollard.11@ucl.ac.uk)
                : Ported from MATLAB implementation by 
                  Gavin C. Cawley and Nicola L. C. Talbot
    Description : Simple python implementation of Sammon's non-linear
                  mapping algorithm [1].
    References  : [1] Sammon, John W. Jr., "A Nonlinear Mapping for Data
                  Structure Analysis", IEEE Transactions on Computers,
                  vol. C-18, no. 5, pp 401-409, May 1969.
    Copyright   : (c) Dr Gavin C. Cawley, November 2007.
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
    """

    # Create distance matrix unless given by parameters
    if inputdist == 'distance':
        D = x
        if init == 'default':
            init = 'cmdscale'
    else:
        D = cdist(x, x)
        if init == 'default':
            init = 'pca'

    if inputdist == 'distance' and init == 'pca':
        raise ValueError("Cannot use init == 'pca' when inputdist == 'distance'")

    if np.count_nonzero(np.diagonal(D)) > 0:
        raise ValueError("The diagonal of the dissimilarity matrix must be zero")

    # Remaining initialisation
    N = x.shape[0]
    scale = 0.5 / D.sum()
    D = D + np.eye(N)

    if np.count_nonzero(D <= 0) > 0:
        raise ValueError("Off-diagonal dissimilarities must be strictly positive")

    Dinv = 1 / D
    if init == 'pca':
        [UU, DD, _] = np.linalg.svd(x)
        y = UU[:, :n] * DD[:n]
    elif init == 'cmdscale':
        y, e = cmdscale(D)
        y = y[:, :n]
    else:
        y = np.random.normal(0.0, 1.0, [N, n])
    one = np.ones([N, n])
    d = cdist(y, y) + np.eye(N)
    dinv = 1. / d
    delta = D - d
    E = ((delta ** 2) * Dinv).sum()

    # Get on with it
    for i in range(maxiter):

        # Compute gradient, Hessian and search direction (note it is actually
        # 1/4 of the gradient and Hessian, but the step size is just the ratio
        # of the gradient and the diagonal of the Hessian so it doesn't
        # matter).
        delta = dinv - Dinv
        deltaone = np.dot(delta, one)
        g = np.dot(delta, y) - (y * deltaone)
        dinv3 = dinv ** 3
        y2 = y ** 2
        H = np.dot(dinv3, y2) - deltaone - np.dot(2, y) * np.dot(dinv3, y) + y2 * np.dot(dinv3, one)
        s = -g.flatten(order='F') / np.abs(H.flatten(order='F'))
        y_old = y

        # Use step-halving procedure to ensure progress is made
        for j in range(maxhalves):
            s_reshape = np.reshape(s, (-1, n), order='F')
            y = y_old + s_reshape
            d = cdist(y, y) + np.eye(N)
            dinv = 1 / d
            delta = D - d
            E_new = ((delta ** 2) * Dinv).sum()
            if E_new < E:
                break
            else:
                s = 0.5 * s

        # Bomb out if too many halving steps are required
        if j == maxhalves - 1:
            if display:
                print('Warning: maxhalves exceeded. Sammon mapping may not converge...')

        # Evaluate termination criterion
        if abs((E - E_new) / E) < tolfun:
            if display:
                print('TolFun exceeded: Optimisation terminated')
            break

        # Report progress
        E = E_new
        if display > 1:
            print('epoch = %d : E = %12.10f' % (i + 1, E * scale))

    if i == maxiter - 1:
        print('Warning: maxiter exceeded. Sammon mapping may not have converged...')

    # Fiddle stress to match the original Sammon paper
    E = E * scale

    return [y, E]




import numpy as np

from gmtb.util import pdist
from gmtb.util.distance import euclidean_dist


# greedy resampling to be used by sparsemap
def greedy_resampling(obj_set, dist_func, emb, p_res, num_dim):

    # if the embedding is already small enough
    if emb.shape[1] <= num_dim:
        return emb

    n_items = len(obj_set)

    # choose random subset
    subset = np.random.permutation(n_items).astype(np.int64)
    subset = subset[:round(n_items * p_res)]
    obj_subset = [obj_set[i] for i in subset]

    # calculate distances of items to the subset
    dist_subset_obj = pdist(obj_subset, dist_func)

    # create finished embedding
    emb_out = np.zeros((n_items,num_dim))

    for d in range(num_dim):

        n_features = emb.shape[1]
        k_stress = np.zeros(n_features)

        # prepare array to attach values
        ft = np.zeros((n_items, d+1))
        ft[:,:d] = emb_out[:,:d]

        #for each possible dimension in the original embedding
        for f in range(n_features):
            # try attaching this dimension to the final embedding
            ft[:,d] = emb[:,f]

            # calculate distance
            ft_dist = pdist(ft[subset,:], euclidean_dist)
            k_stress[f] = np.sum((ft_dist - dist_subset_obj)**2)

        # get the best feature combination
        f_best = np.argmin(k_stress)

        # attach best feature
        emb_out[:, d] = emb[:, f_best]

        # remove from full embedding
        np.delete(emb, f_best, axis=1)

    return emb_out


# this is an implementation of sparsemap published in 'cluster-preserving embedding of
# proteins' (1999) by Hristescu and Farach-Colton.
def sparsemap(obj_set, dist_func, num_dim, dist_mat=None, p_sigma=0.2, p_res=0.2):

    n_items = len(obj_set)

    # number of rows and columns for the reference matrix
    n_sets_rows = np.floor(np.log2(n_items)).astype(np.int64)
    n_sets_cols = np.floor(np.log2(n_items)).astype(np.int64)

    # number of reference sets
    n_sets = n_sets_cols * n_sets_rows

    # reference sets as row vector
    ref_sets = [None] * n_sets

    # create matrix of reference sets R. Every cell contains a set
    x_index = 0

    for i in range(n_sets_rows):
        for j in range(n_sets_cols):
            # create random subset
            x = np.random.permutation(n_items)
            ref_sets[x_index] = x[:2**i]

            # if  the stored set contains all avaliable items
            if len(ref_sets[x_index]) == n_items:
                ref_sets[x_index+1:] = []
                n_sets = x_index+1
                break

            x_index = x_index + 1

    # matrix for lipschitz-embedding
    emb = np.zeros((n_items,n_sets))

    for i in range(n_sets):
        # get current reference set and its length
        x = ref_sets[i]

        # for every item with index p of 0
        for p in range(n_items):
            # for every item q in the reference set, compute the approximated distance
            approx_dist_x = np.zeros(len(x))

            for q in range(len(x)):
                # if  there are embedded features avaliable, the approximated distance will be computed
                if i > 0:
                    approx_dist_x[q] = np.sum((emb[p,:i-1] - emb[x[q],:i-1]) ** 2)
                else:
                    if dist_mat is None:
                        approx_dist_x[q] = dist_func(obj_set[p], obj_set[q])
                    else:
                        approx_dist_x[q] = dist_mat[p,q]

            # if there are embedded features avaliable, the current distances are approximated
            # and a subset of true distances will be calculated
            if i > 0:

                # calcualte order
                approx_dist_x_order = np.argsort(approx_dist_x)

                # choose number of items in current subset for which the true distances will be calculated
                n_true_dist_x_sub = np.ceil(len(approx_dist_x)*p_sigma).astype(np.int64)

                # calculate true distances
                true_dist_x_sub = np.zeros(n_true_dist_x_sub)
                for q in range(n_true_dist_x_sub):
                    if dist_mat is None:
                        true_dist_x_sub[q] = dist_func(obj_set[p],obj_set[x[approx_dist_x_order[q]]])
                    else:
                        true_dist_x_sub[q] = dist_mat[p, x[approx_dist_x_order[q]]]
            else:
                true_dist_x_sub = approx_dist_x

            # the minimum distance is the k-th feature for p
            emb[p,i] = np.min(true_dist_x_sub)

    # perform greedy resampling to reduce dimensions
    return greedy_resampling(obj_set, dist_func, emb, p_res, num_dim)


if __name__ == '__main__':
    # create random points
    points = np.random.rand(10, 10)

    #compute distances
    dist = pdist(points, euclidean_dist)

    res1 = sparsemap(points, euclidean_dist, 5)
    res2 =  sparsemap(points, euclidean_dist, 5, dist_mat=dist)

    dist_res1 = pdist(res1,euclidean_dist)
    dist_res2 = pdist(res2,euclidean_dist)

    print("Difference:")





import numpy as np


# base scalar product of the edit kernel
def edit_prod(x,y,dist_func,origin):
    return -1/2 * (np.power(dist_func(x,y),2) - np.power(dist_func(x,origin),2) - np.power(dist_func(y,origin),2))


# base scalar product of the batch version of the edit kernel
def edit_prod_batch(dist_mat, origin_idx):
    return 1 / 2 * (np.add.outer(np.squeeze(dist_mat[:, origin_idx] * dist_mat[:, origin_idx]),
                                 np.squeeze(dist_mat[origin_idx,:] * dist_mat[origin_idx,:]))
                    - dist_mat * dist_mat)


# normal version of the distance substitution kernel
def edit_kernel(x, y, dist_func, origin_set):
    kernels = np.zeros(len(origin_set));

    for i in range(len(origin_set)):
        kernels[i] = edit_prod(x,y,dist_func,origin_set[i])

    return np.sum(kernels)


def edit_kernel_batch(dist_mat, origin_set_ind):
    kernel_mat = np.zeros((np.shape(dist_mat)[0], np.shape(dist_mat)[1], len(origin_set_ind)))

    for i in range(len(origin_set_ind)):
        kernel_mat[:,:,i] = edit_prod_batch(dist_mat,origin_set_ind[i])

    return np.sum(kernel_mat, axis=2)



import numpy as np


def pdist(set1, func, set2=None):

    if set2 is None:
        n = len(set1)
        p_matrix = np.zeros((n,n))

        for i in range(n):
            for j in range(i, n):
                p_matrix[i, j] = func(set1[i], set1[j])
                p_matrix[j, i] = p_matrix[i, j]

        return p_matrix

    else:
        n = len(set1)
        m = len(set2)

        p_matrix = np.zeros((n, m))

        for i in range(n):
            for j in range(m):
                p_matrix[i, j] = func(set1[i], set2[j])

        return p_matrix.squeeze()



import numpy as np
from munkres import Munkres


# relabel clustering y to match the most amount of labels with x without changing clusterings themselves
def relabel(y, x):

    # compute cost matrix
    max_label = max(max(x), max(y))
    cost_matrix = np.zeros((max_label + 1, max_label + 1))
    for i in range(len(x)):
        cost_matrix[x[i], y[i]] = cost_matrix[x[i], y[i]] - 1

    # assignment
    assignment = Munkres().compute(cost_matrix)

    #assignment = munkres(cost_matrix).nonzero()

    # relabel
    y_tmp = np.copy(y)
    for i in range(1,max_label+1):
        y_tmp[y == assignment[i][1]] = assignment[i][0]

    return y_tmp


# computes the partition distance between two clusterings
def partition_dist(x, y):
    return np.sum(x != relabel(y, x))


# computes the weighted mean between two clusterings based on the partition distance
def partition_wm(x, y, alpha):
    y_tmp = relabel(y, x)

    # compute distance
    d = np.sum(x != y_tmp)

    # find labels that differ
    idx = (x != y_tmp)

    d_changed = 0
    i = 0
    mean = np.copy(x)

    while d_changed < round(alpha*d):
        # find next index
        while not idx[i]:
            i = i + 1
        # change it
        mean[i] = y_tmp[i]
        i = i + 1
        d_changed = d_changed + 1

    return mean



import gmtb.dpe.reconstruction
import gmtb.util as util


# wrapper for the reconstruction methods
# returns:
#   rec_obj - the reconstructed object
#   best_value - SOD of the reconstructed object
def reconstruct(object_vector, embedding, object_set, dist_func, weighted_mean_func, rec_type):

    # TODO check input

    # TODO check embedding

    # if only one object: nothing to do
    if len(object_set) == 1:
        return object_set[0], 0

    # do reconstruction
    if rec_type == "linear":
        return gmtb.dpe.reconstruction.linear(object_vector, embedding, object_set, dist_func, weighted_mean_func, 2)

    elif rec_type == "triangular":
        return gmtb.dpe.reconstruction.linear(object_vector, embedding, object_set, dist_func, weighted_mean_func, 3)

    elif rec_type == "linear_recursive":
        return gmtb.dpe.reconstruction.linear_recursive(object_vector, embedding, object_set, dist_func, weighted_mean_func)

    elif rec_type == "triangular_recursive":
        return gmtb.dpe.reconstruction.triangular_recursive(object_vector, embedding, object_set, dist_func, weighted_mean_func)

    elif rec_type == "linear_search_linear_recursive":
        rec_obj, best_val = gmtb.dpe.reconstruction.triangular_recursive(object_vector, embedding, object_set,
                                                                dist_func, weighted_mean_func)
        return util.linear_search(rec_obj, object_set, dist_func, weighted_mean_func)

    else:
        raise ValueError("reconstruct: rec_type not found")

    # TODO: what to do in case of unsuccessfull reconstruction?



import numpy as np


# This is an implementation of metricmap published in 'An Index Structure
# for Data Mining and Clustering' (2000) by Wang.
#   O:            list of cells which contains the items
#   d:            distance function (O x O -> R)
#   k:            destination dimension of the embedding
def metricmap(obj_set, dist_func, num_dim, dist_mat=None):
    n_items = len(obj_set)

    if num_dim > np.floor(n_items / 2):
        num_dim = np.floor(n_items / 2).astype(np.int64)

    # create subset with double number of elements than dimension
    m = 2 * num_dim
    subset_id = np.random.permutation(n_items)
    subset_id = subset_id[:m]
    subset = [obj_set[i] for i in subset_id]

    # set first item as orgin and remove it
    s0 = subset[0]
    s0_id = subset_id[0]
    subset.pop(0)
    subset_id = subset_id[1:]
    m = m-1

    # compute distances with respect to the origin
    m_mat = np.zeros((m, m))
    for i in range(m-1):
        for j in range(i+1, m):
            # compute distances in the reference set
            if dist_mat is None:
                m_mat[i, j] = (dist_func(subset[i], s0)**2 + dist_func(subset[j], s0)**2
                               - dist_func(subset[i], subset[j])**2) / 2
            else:
                m_mat[i, j] = (dist_mat[subset_id[i], s0_id] ** 2 + dist_mat[subset_id[j], s0_id] ** 2
                               - dist_mat[subset_id[i], subset_id[j]] ** 2) / 2
            m_mat[j, i] = m_mat[i, j]

    # calculate orthogonal basis of m_mat and the eigenvalues
    eig_val, eig_vec = np.linalg.eigh(m_mat)

    eig_val = np.flip(eig_val)
    eig_vec = np.flip(eig_vec, axis=1)
    eig_vec_inv = eig_vec.T

    # get a num_dim submatrix of eig_vec
    eig_vec_inv_dim = eig_vec_inv[:num_dim, :num_dim]

    # make a diagonal matrix with the square root of the corresponding eigenvalues
    c = np.sqrt(np.abs(eig_val[:num_dim]))
    c[c == 0] = 1
    c_mat = np.diag(c)

    # ********* project objects *********
    emb = np.zeros((n_items, num_dim))

    # current projection
    h_k = np.zeros(num_dim)

    for o in range(n_items):
        for d in range(num_dim):
            # compute projection of object_set[o] in dimension d
            if dist_mat is None:
                h_k[d] = (dist_func(s0, obj_set[o])**2 + dist_func(s0, subset[d])**2
                          - dist_func(obj_set[o], subset[d])**2) / 2
            else:
                h_k[d] = (dist_mat[s0_id, o]**2 + dist_mat[s0_id, subset_id[d]]**2
                          - dist_mat[o, subset_id[d]]**2) / 2

        # map object
        emb[o, :] = np.linalg.solve(c_mat, eig_vec_inv_dim) @ h_k

    return emb


if __name__ == '__main__':
    # create random points
    points = np.random.rand(10, 10)

    #compute distances
    from gmtb.util import pdist
    from gmtb.util.distance import euclidean_dist
    dist = pdist(points, euclidean_dist)

    res1 = metricmap(points, euclidean_dist, 5)
    res2 = metricmap(points, euclidean_dist, 5, dist_mat=dist)

    dist_res1 = pdist(res1,euclidean_dist)
    dist_res2 = pdist(res2,euclidean_dist)

    print("Difference:")




import numpy as np

import gmtb.kernel
import gmtb.kernel.reconstruction
import gmtb.util


# computes median using kernel functions.
# paramter dist_func is only needed in case of linear search
def compute_median(object_set, weights, kernel_func, weighted_mean_func, rec_type, kernel_matrix=None, dist_func=None):
    n = len(object_set)

    # compute kernels
    if kernel_matrix is None:
        kernel_matrix = gmtb.util.pdist(object_set, kernel_func)

    if type(weights) is str:
        if weights == "mean":
            weights = np.ones(n)
            dist = np.sqrt(np.sum(kernel_matrix) / np.power(n, 2) - 2/n * np.sum(kernel_matrix, 0) + np.diag(kernel_matrix))
        elif weights == "median":
            weights = gmtb.kernel.compute_weights(kernel_matrix, True)[0]
            dist = 1 / weights
        else:
            raise ValueError("compute_median: only median and mean allowed as string values for weights")
    else:
        #TODO test this
        dist = np.sqrt(np.sum(weights @ np.transpose(weights) * kernel_matrix) / np.power(np.sum(weights, axis=0), 2)
                       - 2 * np.sum(weights * kernel_matrix, axis=0).T / np.sum(weights, axis=0) + np.diag(kernel_matrix))

    # reconstruction
    if rec_type == "linear":
        return gmtb.kernel.reconstruction.linear(object_set, weights, dist, kernel_func, weighted_mean_func, kernel_matrix, 2)
    elif rec_type == "triangular":
        return gmtb.kernel.reconstruction.linear(object_set, weights, dist, kernel_func, weighted_mean_func, kernel_matrix, 3)
    elif rec_type == "linear_recursive":
        return gmtb.kernel.reconstruction.linear_recursive(object_set, weights, dist, kernel_func, weighted_mean_func,
                                               kernel_matrix)
    elif rec_type == "triangular_recursive":
        return gmtb.kernel.reconstruction.triangular_recursive(object_set, weights, dist, kernel_func, weighted_mean_func,
                                                   kernel_matrix)
    elif rec_type == "linear_search_linear_recursive":
        obj, bv = gmtb.kernel.reconstruction.linear_recursive(object_set, weights, dist, kernel_func, weighted_mean_func,
                                                  kernel_matrix)
        return gmtb.util.linear_search(obj, object_set, dist_func, weighted_mean_func)
    else:
        raise NotImplementedError("kernel.compute_median: reconstruction method not found")





from setuptools import setup, find_packages, Extension
import numpy

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

with open('REAMDE.md',encoding='utf-8') as f:
    readme = f.read()

ext = [
      Extension(name="gmtb.util.distance.kendall_tau_dist_cy",
                sources=["gmtb/util/distance/kendall_tau_dist_cy.pyx"],
                include_dirs=[numpy.get_include()])
]

setup(name='gmtb',
      install_requirements=requirements,
      setup_requires=[
          'setuptools>=18.0',
          'cython>=0.28.4',
      ],
      version='1.0.1',
      description='Python version of the Generalized Median Toolbox.',
      long_description_content_type='text/markdown',
      long_description=readme,
      author='Andreas Nienkötter',
      author_email='a.nienkoetter@uni-muenster.de',
      url='https://www.uni-muenster.de/PRIA/en/forschung/dpe.html',
      license='MIT',
      packages=find_packages(exclude=('test', 'docs')),
      ext_modules=ext,
      classifiers=[
          "License :: OSI Approved :: MIT License",
          "Programming Language :: Python :: 3",
      ]
      )



from .compute_alpha import compute_alpha
from .compute_median import compute_median
from .compute_median_vector import compute_median_vector
from .best_weighted_mean import best_weighted_mean
from .reconstruct import reconstruct
from .embed import embed



import numpy as np


def compute_median_vector(vector_matrix, threshold=1e-10, max_iter=1000, max_iter_restart=3):

    dim = np.shape(vector_matrix)[1]
    y_old = np.zeros(dim)
    y_new = np.sum(vector_matrix, axis=0)/dim

    iter_restart = 0
    iter = 0

    # change initialization slightly as long as it lies on one of the points
    diff = np.sum(np.power(vector_matrix - y_new[np.newaxis, :], 2), axis=1)
    while np.any(diff == 0):
        y_new = y_new + 0.001
        diff = np.sum(vector_matrix - y_new[np.newaxis, :])

    while np.sum(np.abs(y_new - y_old)) > threshold and iter < max_iter:
        iter = iter + 1
        y_old = y_new

        # stop when the iteration coincidents with one point
        diff = np.sum(np.power(vector_matrix - y_old[np.newaxis, :], 2),axis=1)
        if np.any(diff == 0):
            return y_old

        eukl_dist = np.sqrt(diff)
        numerator = np.sum(vector_matrix / eukl_dist[:, np.newaxis], axis=0)
        denominator = np.sum(1 / eukl_dist)
        y_new = numerator / denominator

    return y_new



#!/usr/bin/env python

# This implements the FastMap algorithm
# for mapping points where only the distance between them is known
# to N-dimension coordinates.
# The FastMap algorithm was published in:
#
# FastMap: a fast algorithm for indexing, data-mining and
# visualization of traditional and multimedia datasets
# by Christos Faloutsos and King-Ip Lin
# http://portal.acm.org/citation.cfm?id=223812

# This code made available under the BSD license,
# details at the bottom of the file
# Copyright (c) 2009, Gunnar Aastrand Grimnes

import math
import random

# need scipy as usual
import scipy

# we will repeat the pick-pivot points heuristic this many times
# a higher value means "better" results, but 1 also works well
DISTANCE_ITERATIONS = 1


class FastMap:

    def __init__(self, dist, verbose=False):
        if dist.max() > 1:
            dist /= dist.max()

        self.dist = dist
        self.verbose = verbose

    def _furthest(self, o):
        mx = -1000000
        idx = -1
        for i in range(len(self.dist)):
            d = self._dist(i, o, self.col)
            if d > mx:
                mx = d
                idx = i

        return idx

    def _pickPivot(self):
        """Find the two most distant points"""
        o1 = random.randint(0, len(self.dist) - 1)
        o2 = -1

        i = DISTANCE_ITERATIONS

        while i > 0:
            o = self._furthest(o1)
            if o == o2: break
            o2 = o
            o = self._furthest(o2)
            if o == o1: break
            o1 = o
            i -= 1

        self.pivots[self.col] = (o1, o2)
        return (o1, o2)

    def _map(self, K):
        if K == 0: return

        px, py = self._pickPivot()

        if self.verbose: print
        "Picked %d,%d at K=%d" % (px, py, K)

        if self._dist(px, py, self.col) == 0:
            return
        for i in range(len(self.dist)):
            self.res[i][self.col] = self._x(i, px, py)

        self.col += 1
        self._map(K - 1)

    def _x(self, i, x, y):
        """Project the i'th point onto the line defined by x and y"""
        dix = self._dist(i, x, self.col)
        diy = self._dist(i, y, self.col)
        dxy = self._dist(x, y, self.col)
        return (dix + dxy - diy) / 2 * math.sqrt(dxy)

    def _dist(self, x, y, k):
        """Recursively compute the distance based on previous projections"""
        if k == 0: return self.dist[x, y] ** 2

        rec = self._dist(x, y, k - 1)
        resd = (self.res[x][k] - self.res[y][k]) ** 2
        return rec - resd

    def map(self, K):
        self.col = 0
        self.res = scipy.zeros((len(self.dist), K))
        self.pivots = scipy.zeros((K, 2), "i")
        self._map(K)
        return self.res


def fastmap(dist, K):
    """dist is a NxN distance matrix
    returns coordinates for each N in K dimensions
    """

    return FastMap(dist, True).map(K)


# Below here are methods for testing

def vlen(x, y):
    return math.sqrt(sum((x - y) ** 2))


def distmatrix(p, c=vlen):
    dist = scipy.zeros((len(p), len(p)))
    for x in range(len(p)):
        for y in range(x, len(p)):
            if x == y: continue
            dist[x, y] = c(p[x], p[y])
            dist[y, x] = dist[x, y]

    return dist


def distortion(d1, d2):
    return scipy.sum(((d1 / d1.max()) - (d2 / d2.max())) ** 2) / d1.size


def distortiontest():
    points = []

    n = 10
    mean = 10
    dim = 5

    print
    "Generating %d %d-D points randomly distributed between [0-%d]" % (n, dim, mean)
    for i in range(n):
        points.append(scipy.array([random.uniform(0, mean) for x in range(dim)]))
    print
    "Computing distance matrix"
    dist = distmatrix(points)

    print
    "Mapping"

    p1 = fastmap(dist, 1)
    print
    "K=1"
    print
    "Distortion: ", distortion(distmatrix(p1), dist)

    p2 = fastmap(dist, 2)
    print
    "K=2"
    print
    "Distortion: ", distortion(distmatrix(p1), dist)

    p3 = fastmap(dist, 3)
    print
    "K=3"
    print
    "Distortion: ", distortion(distmatrix(p3), dist)

    import pylab
    pylab.scatter([x[0] / mean for x in points], [x[1] / mean for x in points], s=50)
    pylab.scatter([x[0] for x in p2], [x[1] for x in p2], c="r")
    pylab.show()


# Copyright (c) 2009, Gunnar Aastrand Grimnes
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

#     * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
#     * Neither the name of the <ORGANIZATION> nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



import numpy as np
from ..compute_alpha import compute_alpha
from ..best_weighted_mean import best_weighted_mean


# implementation of the triangular recursive reconstruction method
# returns:
#   rec_obj - reconstructed object
#   best_value - value of the reconstructed object (at the moment SOD)
def triangular_recursive(object_set,weights,dist,kernel_func,weighted_mean_func,kernel_matrix):

    # return if only one or two objects
    if len(object_set) == 1:
        return object_set[0], 0

    # save original set
    orig_set = object_set.copy()
    object_set = object_set.copy()

    # arrays for intermediate kernel values
    k_mean_set = kernel_matrix
    k_mean_mean = np.diag(kernel_matrix)

    # first best value: set median
    ind = np.argsort(dist)
    best_value = np.sum(np.sqrt(k_mean_mean[ind[0]] - 2 * k_mean_set[ind[0], :] + np.diag(kernel_matrix)))
    new_mean = object_set[ind[0]]

    # precompute k_median (kernel of the true median with itself)
    k_median_vec = np.sum(np.outer(weights,weights) * kernel_matrix) / np.sum(weights)**2

    while len(object_set) > 1:

        ind = np.argsort(dist)

        new_object_set = []
        new_k_mean_set = []
        new_k_mean_mean = []
        new_dist = []

        # for every three objects
        for i in range(0, len(object_set)-2, 3):

            # compute median between first two
            alpha = compute_alpha(weights, k_mean_set[ind[i]], k_mean_mean[ind[i]], k_mean_set[ind[i+1]], k_mean_mean[ind[i+1]],
                                  kernel_func(object_set[ind[i]],object_set[ind[i+1]]))
            nm, kms, kmm, bv = \
                best_weighted_mean(object_set[ind[i]], object_set[ind[i+1]], alpha, orig_set, kernel_matrix, kernel_func,
                                   weighted_mean_func)

            # compute median between result and third
            alpha = compute_alpha(weights, kms, kmm, k_mean_set[ind[i + 2]], k_mean_mean[ind[i + 2]],
                                  kernel_func(nm, object_set[ind[i + 2]]))
            nm, kms, kmm, bv = \
                best_weighted_mean(nm, object_set[ind[i + 2]], alpha, orig_set, kernel_matrix, kernel_func,
                                   weighted_mean_func)

            new_object_set.append(nm)
            new_k_mean_set.append(kms)
            new_k_mean_mean.append(kmm)
            tmp_dist = (k_median_vec
                                    - 2 * np.sum(weights * kms) / np.sum(weights)
                                    + kmm)
            if tmp_dist < 0:
                tmp_dist = (dist[ind[i]] + dist[ind[i+1]] + dist[ind[i+2]])/3
            else:
                tmp_dist = np.sqrt(tmp_dist)
            new_dist.append(tmp_dist)

            # save best
            if bv < best_value:
                best_value = bv
                new_mean = nm



        # if only one leftover object: just copy
        if np.remainder(len(object_set), 3) == 1:
            new_object_set.append(object_set[-1])
            new_k_mean_mean.append(k_mean_mean[-1])
            new_k_mean_set.append(k_mean_set[-1])
            new_dist.append(dist[-1])

        # else: linear
        elif np.remainder(len(object_set), 3) == 2:
            alpha = compute_alpha(weights, k_mean_set[ind[-2]], k_mean_mean[ind[-2]], k_mean_set[ind[-1]],
                                  k_mean_mean[ind[-1]],
                                  kernel_func(object_set[ind[-2]], object_set[ind[-1]]))

            nm, kms, kmm, bv = \
                best_weighted_mean(object_set[ind[-2]], object_set[ind[-1]], alpha, orig_set, kernel_matrix,
                                   kernel_func,
                                   weighted_mean_func)

            new_object_set.append(nm)
            new_k_mean_set.append(kms)
            new_k_mean_mean.append(kmm)
            tmp_dist = (k_median_vec
                - 2 * np.sum(weights * kms) / np.sum(weights)
                + kmm)
            if tmp_dist < 0:
                tmp_dist = (dist[ind[-2]] + dist[ind[-1]]) / 2
            else:
                tmp_dist = np.sqrt(tmp_dist)
            new_dist.append(tmp_dist)

            if bv < best_value:
                best_value = bv
                new_mean = nm

        # prepare next iteration
        object_set = new_object_set
        k_mean_set = new_k_mean_set
        k_mean_mean = new_k_mean_mean
        dist = new_dist

    # end: return best result
    return new_mean, best_value




import numpy as np

# for testing
import gmtb.util
import gmtb.util.distance


# base scalar product of the distance substitution kernel
def dist_sub_prod(x,y,dist_func,origin):
    return -1/2 * (np.power(dist_func(x,y),2) - np.power(dist_func(x,origin),2) - np.power(dist_func(y,origin),2))


# normal version of the distance substitution kernel
def dist_sub(x, y, dist_func, kernel_type, origin=None, param=None):
    if kernel_type == "lin":
        return dist_sub_prod(x,y,dist_func,origin)

    if kernel_type == "nd":
        return -np.power(dist_func(x, y), param)

    if kernel_type == "pol":
        return np.power(1 + param[0] * dist_sub_prod(x,y,dist_func,origin), param[1])

    if kernel_type == "rbf":
        return np.exp(-param * np.power(dist_func(x,y),2))
    else:
        raise ValueError("kernel_type not found")


# base scalar product of the batch version of the distance substitution kernel
def dist_sub_batch_prod(dist_mat, origin_idx):
    return 1 / 2 * (np.add.outer(np.squeeze(dist_mat[:, origin_idx] * dist_mat[:, origin_idx]),
                                 np.squeeze(dist_mat[origin_idx,:] * dist_mat[origin_idx,:]))
                    - dist_mat * dist_mat)


# batch version of the distance substitution kernels
def dist_sub_batch(dist_mat, kernel_type, origin_idx=None, param=None):
    if kernel_type == "lin":
        return dist_sub_batch_prod(dist_mat, origin_idx)
    if kernel_type == "nd":
        return -np.power(dist_mat, param)
    if kernel_type == "pol":
        return np.power(1 + param[0] * dist_sub_batch_prod(dist_mat,origin_idx), param[1])
    if kernel_type == "rbf":
        return np.exp(-param * dist_mat * dist_mat)
    else:
        raise ValueError("kernel_type not found")



if __name__ == "__main__":


    points = np.random.rand(10,5)

    def kernel(x,y):
        return dist_sub(x,y,gmtb.util.distance.euclidean_dist,"lin",origin=points[0])

    dist_mat = gmtb.util.pdist(points, gmtb.util.distance.euclidean_dist)
    kernel_mat = gmtb.util.pdist(points,kernel)
    kernel_mat_batch = dist_sub_batch(dist_mat, "lin", origin_idx=0)

    print("Max diff: %f" % np.max(np.abs(kernel_mat - kernel_mat_batch)))